import pytest


def test_without(testdir):
    testdir.makepyfile(
        """
        def test_01():
            a = 5
            b = 10
            assert a + b == 15
        def test_02():
            a = 10
            b = 5
            assert a + b != 15
        """
    )
    result = testdir.runpytest()
    result.assert_outcomes(passed=1, failed=1)


def test_help(testdir):
    testdir.makepyfile(
        """
        def test_01():
            a = "pytest"
            b = "py"
            assert a == b
        """
    )
    result = testdir.runpytest("--help")
    result.stdout.fnmatch_lines(["*--concurrent*"])
    result.stdout.fnmatch_lines(["*--sub_group=*"])


def test_on(testdir):
    testdir.makepyfile(
        """
        import pytest
        @pytest.mark.parametrize("group", ["group_1", "group_2", "group_3"])
        def test_01(group):
            a = "hello world"
            b = "hello world"
            assert a == b
            
        def test_02():
            a = "hello"
            b = "world"
            assert a == b
        """
    )
    result = testdir.runpytest("-svv", "--concurrent=on")
    result.assert_outcomes(passed=3, failed=1)


def test_on_group_none_param(testdir):
    testdir.makepyfile(
        """
        def test_01():
            a = "hello world"
            b = "hello world"
            assert a == b

        def test_02():
            a = "hello"
            b = "world"
            assert a == b
        """
    )
    result = testdir.runpytest("--concurrent=on", "--sub_group=name")
    result.assert_outcomes(passed=1, failed=1)


def test_on_group_have_param(testdir):
    testdir.makepyfile(
        """
        import pytest
        @pytest.mark.parametrize("name", ["group_1", "group_2", "group_3"])
        def test_01(name):
            a = "hello world"
            b = "hello world"
            assert a == b

        @pytest.mark.parametrize("name", ["group_1", "group_2", "group_3"])
        def test_02(name):
            a = "hello"
            b = "world"
            assert a == b

        @pytest.mark.parametrize("name", ["group_1", "group_2", "group_3"])
        def test_03(name):
            a = "hello"
            b = "world"
            assert a == b

        @pytest.mark.parametrize("name", ["group_1", "group_2", "group_3"])
        def test_04(name):
            a = "hello"
            b = "world"
            assert a == b

        @pytest.mark.parametrize("name", ["group_1", "group_2", "group_3"])
        def test_05(name):
            a = "hello"
            b = "world"
            assert a == b
        """
    )
    result = testdir.runpytest("-svv", "--concurrent=on", "--sub_group=name")
    result.assert_outcomes(passed=3, failed=12)


def test_off(testdir):
    testdir.makepyfile(
        """
        def test_01():
            a = "hello world"
            b = "hello world"
            assert a == b

        def test_02():
            a = "hello"
            b = "world"
            assert a == b
        """
    )
    result = testdir.runpytest("--concurrent=off")
    result.assert_outcomes(passed=1, failed=1)


def test_off_group(testdir):
    testdir.makepyfile(
        """
        import pytest
        @pytest.mark.parametrize("group_by", ["group_1", "group_2", "group_3"])
        def test_01(group_by):
            a = "hello world"
            b = "hello world"
            assert a == b

        def test_02():
            a = "hello"
            b = "world"
            assert a == b
        """
    )
    result = testdir.runpytest("--concurrent=off", "--sub_group=group_by")
    result.assert_outcomes(passed=3, failed=1)


def test_off_group_have_param(testdir):
    testdir.makepyfile(
        """
        import pytest
        @pytest.mark.parametrize("name", ["group_1", "group_2", "group_3"])
        def test_01(name):
            a = "hello world"
            b = "hello world"
            assert a == b

        @pytest.mark.parametrize("name", ["group_1", "group_2", "group_3"])
        def test_02(name):
            a = "hello"
            b = "world"
            assert a == b

        @pytest.mark.parametrize("name", ["group_1", "group_2", "group_3"])
        def test_03(name):
            a = "hello"
            b = "world"
            assert a == b

        @pytest.mark.parametrize("name", ["group_1", "group_2", "group_3"])
        def test_04(name):
            a = "hello"
            b = "world"
            assert a == b

        @pytest.mark.parametrize("name", ["group_1", "group_2", "group_3"])
        def test_05(name):
            a = "hello"
            b = "world"
            assert a == b
        """
    )
    result = testdir.runpytest("--concurrent=off", "--sub_group=name")
    result.assert_outcomes(passed=3, failed=12)


if __name__ == '__main__':
    # pytest.main(["--concurrent=on"])
    # pytest.main(["-svv"])
    pytest.main(["-sv", "test_plugin.py::test_on_group_have_param"])



from setuptools import setup

"""
author: fungaegis
github: https://github.com/fungaegis/pytest_custom_Concurrency
"""

setup(
    name='pytest-custom-concurrency',
    url='https://github.com/fungaegis/fungaegis-pytest-custom-concurrency',
    version='0.0.1',
    author="fungaegis",
    author_email="fungaegis@gmail.com",
    description='Custom grouping concurrence for pytest',
    long_description='Custom grouping concurrence for pytest;\n'
                     ' Usage: cmd line or main function --group=group_name --concurrent=on',
    classifiers=[
        'Framework :: Pytest',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Testing',
        'Programming Language :: Python :: 3.7',
    ],
    license='MIT License',
    py_modules=['pytest_custom_concurrency'],
    keywords=[
        'pytest', 'py.test', 'pytest_custom_concurrency', 'concurrency'
    ],

    install_requires=[
        'pytest'
    ],
    entry_points={
        'pytest11': [
            'custom_concurrency = pytest_custom_concurrency',
        ]
    }
)



from threading import Thread, Lock


def pytest_addoption(parser):
    group = parser.getgroup("custom", "Custom grouping concurrence for pytest")
    group.addoption(
        "--concurrent",
        action="store",
        default="off",
        choices=['on', 'off'],
        help="Default 'off' for switch, option: on or off"
    )
    group.addoption(
        "--sub_group",
        action="store",
        default="group",
        type=str,
        help="Default group name 'group', Can be specified manually"
    )


def pytest_runtestloop(session) -> bool:
    if getattr(session.config.option, "concurrent", None) == "on":
        if session.testsfailed and not session.config.option.continue_on_collection_errors:
            raise session.Interrupted(
                "%d error%s during collection"
                % (session.testsfailed, "s" if session.testsfailed != 1 else "")
            )

        if session.config.option.collectonly:
            return True
        case_obj = {}
        GROUP_NAME = getattr(session.config.option, "sub_group", "group")
        for item in session.items:
            if getattr(item, "callspec", None):
                param = item.callspec.params
                for k, y in param.items():
                    if k == GROUP_NAME:
                        case_obj.setdefault(y, list()).append(item)
            else:
                case_obj.setdefault("default", list()).append(item)

        lock = Lock()
        def run(items):
            for i, item in enumerate(items):
                nextitem = items[i + 1] if i + 1 < len(items) else None
                item.config.hook.pytest_runtest_protocol(item=item, nextitem=nextitem)
                lock.acquire(timeout=2)
                if session.shouldfail:
                    raise session.Failed(session.shouldfail)
                if session.shouldstop:
                    raise session.Interrupted(session.shouldstop)
                lock.release()

        thread_obj = {}
        for k, y in case_obj.items():
            thread = Thread(target=run, kwargs={"items": y})
            thread_obj[k] = thread
            thread.start()
        for obj in thread_obj.values():
            obj.join()
        return True
